#include "format.hpp"
#include "panic.hpp"
#include "print.hpp"
#include "ranges/enumerate.hpp"
#include "ranges/first.hpp"
#include "ranges/generated.hpp"
#include <cstdlib>
#include <iostream>
#include <stdio.h>

#define UTIL_ASSERT_FORMAT
#include "assert.hpp"
#include "ranges/collect.hpp"
#include <map>
#include <ranges>
#include <set>
#include <string_view>
#include <vector>

using namespace std::string_view_literals;

#include "iterator.hpp"

auto test_iterator_t() {
  int *ptr = nullptr;
  int arr[5] = {0, 1, 2, 3, 4};
  using int_ptr_iterator_t = util::iterator_t<int *>;
  auto a = int_ptr_iterator_t(arr);

  const auto three = a[3];
  a += 2;
  *a = 6;

  for (auto &&e : arr) {
    util::println("{}", e);
  }
}

auto test_enumerate_view() {
  auto enumerated = util::views::from([] { return 1; }) |
                    util::views::enumerate | std::views::take(10);

  auto b = std::ranges::begin(enumerated);

  for (auto &&[i, e] : enumerated) {
    util::println("[{}] {}", i, e);
  }

  for (auto &&[i, e] : enumerated | util::views::enumerate) {
    util::println("[{}] {}", i, e.value);
  }

  auto transformed_enumerate =
      util::views::from([] { return 1; }) |
      std::views::transform([](auto &&i) { return i + 1; }) |
      util::views::enumerate | std::views::take(10);

  for (auto &&[i, e] : transformed_enumerate) {
    util::println("[{}] {}", i, e);
  }
}

auto test_generated_view() {
  {
    const auto view = util::views::from([] { return 's'; });
    auto it = std::ranges::begin(view);

    util::assert_eq(*it, *it);
    util::assert_eq(*it, 's');
  }

  {
    auto count = 0;
    const auto view = util::views::from([&] { return count++; });
    auto it = std::ranges::begin(view);

    util::assert_eq(*it, *it);
    util::assert_ne(*it++, *it++);
    util::assert_ne(*it + 1, *++it);
  }
}

auto test_collect() {
  auto vec = std::vector<int>({1, 2, 3, 4, 5});
  auto range = vec | std::views::filter([](auto &i) { return i % 2; });

  const auto set = util::collect<std::set<int>>(range);
}

auto test_first() {
  const auto random =
      util::views::from([] {
        return "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXY"
            [std::rand() %
             std::strlen(
                 "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXY")];
      }) |
      std::views::take(10) | util::views::collect<std::string>;

  const auto first = random | util::views::first;
  util::println("{}", first);
}

auto test_assert_no_fail() {
  util::assert_ne("asdf"sv, "nsdf"sv);
  util::assert_eq(1, 1);
  util::assert_eq(1 == 1, true);
  util::assert_eq(0, false);
}

auto test_panic() { util::panic("asdf {}", 42); }

int main(int argc, const char *argv[]) {
  printf("Hello, Alloy!\n");

  test_iterator_t();
  test_enumerate_view();
  test_collect();
  test_generated_view();
  test_first();
  test_assert_no_fail();

  util::print("hello {}\n", "world");
  return 1;
}